# Adapter for sklearn
from Dataset import *
from Classifier import *

from sklearn import svm

class sklSVC(Classifier):
    def __init__(self, dataset, configuration):
        Classifier.__init__(self,dataset)
        self.clf = svm.SVC()

    # === Learning ===
    def learn(self):
        self.clf.fit(self.dataset.X, self.dataset.y)
        pass

    def predict(self):
        predictions = self.clf.predict(self.dataset.testX)
        for idx, val in enumerate(predictions):
            self.dataset.test[idx].prediction = val



from Sample import *
from utils import getType
import csv
import numpy as np
import random
import re

np.seterr(divide='ignore', invalid='ignore')

SEED = 123
FOLDS = 5
random.seed(SEED)

# DATASET


class Dataset:

    def __init__(self, filename, resample=0):
        # Load db
        wisepath = re.findall(r"[\w']+", filename)
        self.db_name = wisepath[len(wisepath) - 2]
        self.source_samples = []
        self.test = []
        self.has_header = False
        self.header = []
        self.confusion_matrix = []

        self.classes = {}
        with open(filename, 'rb') as file:
            csvDataset = csv.reader(file, delimiter=',')
            for row in csvDataset:
                if getType(row[0]) == str:
                    self.hasHeader = True
                    self.header = row
                else:
                    label = row[-1]
                    features = row[0:-1]
                    if not label in self.classes:
                        self.classes.update({label: len(self.classes)})
                    self.source_samples.append(Sample(features, self.classes[label]))

        # normalize
        self.normalize()

        # dumb resampling
        if resample:
            random.shuffle(self.source_samples)
            self.source_samples = self.source_samples[0:resample]

        # Count features
        self.features = len(self.source_samples[0].features)

        # Initialize supports
        self.clearSupports()

        # Set base sample set
        self.samples = self.source_samples

        self.prepareCV()

        # Scipy
        self.X = [x.features for x in self.samples]
        self.y = [x.label for x in self.samples]
        self.testX = []

    def prepareCV(self):
        # Prepare CV
        indexes = range(0, len(self.source_samples))
        random.shuffle(indexes)

        self.cv = []
        for index, value in enumerate(indexes):
            self.cv.append((value, index % FOLDS))


    def clearSupports(self):
        for sample in self.source_samples:
            sample.support = np.zeros(len(self.classes))

    def __str__(self):
        return "%s dataset" % (self.db_name)

    def setCV(self,fold):
        self.samples = []
        self.test = []
        for pair in self.cv:
            if pair[1] == fold:
                self.samples.append(self.source_samples[pair[0]])
            else:
                self.test.append(self.source_samples[pair[0]]);
        # SKL
        self.X = [x.features for x in self.samples]
        self.y = [x.label for x in self.samples]
        self.testX = [x.features for x in self.test]

    def score(self):
        self.confusion_matrix = np.zeros((len(self.classes),len(self.classes))).astype(int)

        for sample in self.test:
            self.confusion_matrix[sample.label,sample.prediction] += 1

        true_positives = np.zeros(len(self.classes)).astype(float)
        false_negatives = np.zeros(len(self.classes)).astype(float)
        false_positives = np.zeros(len(self.classes)).astype(float)
        true_negatives = np.zeros(len(self.classes)).astype(float)

        for pro in xrange(0,len(self.classes)):
            true_positives[pro] += self.confusion_matrix[pro,pro]
            for contra in xrange(0,len(self.classes)):
                if pro == contra: continue
                false_negatives[pro] += self.confusion_matrix[pro,contra]
                false_positives[pro] += self.confusion_matrix[contra,pro]
            true_negatives[pro] = sum(sum(self.confusion_matrix)) - true_positives[pro] - false_negatives[pro] - false_positives[pro]

        sensitivity = true_positives / (true_positives + false_negatives)
        specificity = true_negatives / (false_positives + true_negatives)
        ppv = true_positives / (true_positives + false_positives)
        npv = true_negatives / (true_negatives + false_negatives)
        bac = (sensitivity + specificity) / 2

        acc = sum(true_positives + true_negatives) / sum(true_positives + true_negatives + false_positives + false_negatives)

        scores = \
        {
            'sensitivity': sensitivity.mean(),
            'specificity': specificity.mean(),
            'ppv': ppv.mean(),
            'npv': npv.mean(),
            'accuracy': acc,
            'bac': bac.mean()
        }

        return scores

    def normalize(self):
        example = np.array(self.source_samples[0].features)
        # Check if there are any NaN's
        for index, value in enumerate(example):
            if np.isnan(value):
                for sample in self.source_samples:
                    if not np.isnan(sample.features[index]):
                        example[index] = sample.features[index]
                        break

        minimum = np.array(list(example))
        maximum = np.array(list(example))

        for index, sample in enumerate(self.source_samples):
            for index, value in enumerate(sample.features):
                if value < minimum[index]:
                    minimum[index] = value
                if value > maximum[index]:
                    maximum[index] = value

        foo = maximum - minimum

        # O tu sobie poradzmy z 0/0
        for index, value in enumerate(foo):
            if value == 0:
                foo[index] = 1

        for sample in self.source_samples:
            for index, feature in enumerate(sample.features):
                if not np.isnan(feature):
                    normalizedFeature = (feature - minimum[index]) / foo[index]
                    sample.features[index] = normalizedFeature



from abc import ABCMeta, abstractmethod
import numpy as np
import random

# To ensure deterministic loop we use randomization with fixed seed.
SEED = 123
random.seed(SEED)

class Ensemble:
    __metaclass__ = ABCMeta
    def __init__(self, dataset):
        # we're gathering the dataset
        self.dataset = dataset



    def quickLoop(self):
        folds = xrange(0,5,1)
        acc = []
        bac = []
        for fold in folds:
            self.dataset.setCV(fold)
            self.learn()
            self.dataset.clearSupports()
            self.predict()
            scores = self.dataset.score()
            acc.append(scores['accuracy'])
            bac.append(scores['bac'])
        acc = np.mean(acc)
        bac = np.mean(bac)
        return {'acc': acc, 'bac': bac}

    @abstractmethod
    def predict():
        pass

    @abstractmethod
    def learn():
        pass



# IMPORTS
from utils import getType
import csv
import numpy as np
import random

# SAMPLE
class Sample:
	def __init__(self,features,label):
		# Label is INT
		# Feature vector is float
  		width = len(features)
  		self.label = label
  		self.prediction = 0
  		self.support = None

		# Missing values are None
  		for index, value in enumerate(features):
  			if value == '?':
  				features[index] = None

  		self.features = np.array(features).astype(np.float)

	def decidePrediction(self):
  		self.prediction = np.argmax(self.support)



from .Dataset import *
from .Sample import *
from .Classifier import *
from .Ensemble import *
from .KNN import *
from .sklKNN import *
from .sklGNB import *
from .sklDTC import *
from .sklMLP import *
from .sklSVC import *
from .gKNN import *



from Dataset import *
from Classifier import *

from enum import Enum
import numpy as np
import math
import operator
import png
import functools
import colorsys


class gKNN(Classifier):

    def __init__(self, dataset, configuration):
        Classifier.__init__(self, dataset)
        self.k = configuration['k']
        self.g = configuration['g']
        self.d = configuration['d']
        for sample in dataset.source_samples:
            sample.cells = self.featuresToCells(sample.features)

    def featuresToCells(self, features):
        return [int(v*self.g) for v in features]

    # === Learning ===
    def learn(self):
        pass

    def predict(self):
        for sample in self.dataset.test:
            winners = []
            entry_threshold = 2
            cells = self.featuresToCells(sample.features)

            for source in self.dataset.samples:
                cmpCells = self.featuresToCells(source.features)
                comparison = map(abs, (map(operator.sub, cells, cmpCells)))
                shallNotPass = max(comparison) > self.d

                if shallNotPass:
                    continue

                # print "%s vs %s is %s [%i]" % (
                #     str(cells), str(cmpCells), str(comparison), shallNotPass)
                distance = np.linalg.norm(
                    sample.features - source.features)
                if distance < entry_threshold:
                    winners.append((source, distance))

                if len(winners) > self.k:
                    winners.sort(key=lambda distance: distance[1])
                    winners = winners[0:self.k]
                    entry_threshold = winners[-1][1]

            for winner in winners:
                sample.support[winner[0].label] += 1

            sample.decidePrediction()
